import copy
import getopt
import multiprocessing as mp
import os
import sys
import uuid
from datetime import date, datetime
from typing import Dict, List

import parsedatetime.parsedatetime as pdt
import pygit2

from liualgotrader.backtesting.optimizer import backtest_iteration
from liualgotrader.common import config
from liualgotrader.common.hyperparameter import Hyperparameters, Parameter
from liualgotrader.common.tlog import tlog


def motd(filename: str, version: str) -> None:
    """Display welcome message"""

    tlog("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
    tlog("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
    tlog(f"{filename} {version} starting")
    tlog("----------------------------------------------------------")
    tlog("----------------------------------------------------------")


def show_usage():
    print(
        f"usage:\n{sys.argv[0]}  [--concurrency=<int> DEFAULT 4]",
    )


def create_parameters(parameters: Dict) -> List[Parameter]:
    params = []

    for strategy in parameters["strategies"]:
        for name in parameters["strategies"][strategy]:
            params.append(
                Parameter(
                    f"strategies.{strategy}.{name}",
                    **parameters["strategies"][strategy][name],
                )
            )

    tlog(f"created {len(params)} parameters")
    return params


def create_hyperparameters(parameters: Dict) -> Hyperparameters:
    params = create_parameters(parameters)
    tlog(f"created {len(params)} hyper-parameters")
    return Hyperparameters(params)


def dateFromString(s: str) -> date:
    c = pdt.Calendar()
    result, what = c.parse(s)

    dt = None

    # what was returned (see http://code-bear.com/code/parsedatetime/docs/)
    # 0 = failed to parse
    # 1 = date (with current time, as a struct_time)
    # 2 = time (with current date, as a struct_time)
    # 3 = datetime
    if what in (1, 2):
        # result is struct_time
        dt = datetime(*result[:6]).date()
    elif what == 3:
        # result is a datetime
        dt = result.date()

    if dt is None:
        raise ValueError("Don't understand date '" + s + "'")

    return dt


def realize(config: Dict, parameters: tuple) -> Dict:
    d = copy.deepcopy(config)

    for parameter in parameters:
        name = parameter[0].split(".")

        element: Dict = d
        for i in name[:-1]:
            element = element[i]
        element[name[-1]] = parameter[1]

    return d


def optimize(
    start_date: date,
    end_date: date,
    hyperparameters: Hyperparameters,
    parameters: List[Parameter],
    conf_dict: Dict,
    concurrency: int,
):
    optimizer_session_id = str(uuid.uuid4())
    tlog("-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-")
    tlog(f"optimizer session_id {optimizer_session_id} starting")
    tlog("-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-")

    processes = []
    for hyperparameter in iter(hyperparameters):
        parameters_values = (x() for x in parameters)
        p = mp.Process(
            target=backtest_iteration,
            args=(
                optimizer_session_id,
                start_date,
                end_date,
                realize(realize(conf_dict, hyperparameter), parameters_values),  # type: ignore
                hyperparameter,
            ),
        )
        processes.append(p)
        p.start()

        if len(processes) == concurrency:
            for p in processes:
                p.join()
            processes = []

    print("done.")


def load_configuration(filename: str):
    try:
        import toml

        conf_dict = toml.load(filename)
        tlog(f"loaded configuration file from {filename}")
    except FileNotFoundError:
        tlog(f"[ERROR] could not locate tradeplan file {filename}")
        sys.exit(0)

    return conf_dict


def main_cli() -> None:
    mp.set_start_method("spawn")

    if len(sys.argv) > 2:
        show_usage()
        sys.exit(0)

    concurrency: int = 4
    try:
        opts, args = getopt.getopt(
            sys.argv[1:],
            "c",
            ["concurrency="],
        )
        for opt, arg in opts:
            if opt in ("--concurrency", "-c"):
                concurrency = int(arg)
    except getopt.GetoptError as e:
        print(f"Error parsing options:{e}\n")
        show_usage()
        sys.exit(0)
    try:
        config.build_label = pygit2.Repository("../").describe(
            describe_strategy=pygit2.GIT_DESCRIBE_TAGS
        )
    except pygit2.GitError:
        import liualgotrader

        config.build_label = liualgotrader.__version__ if hasattr(liualgotrader, "__version__") else ""  # type: ignore

    config.filename = os.path.basename(__file__)
    motd(filename=config.filename, version=config.build_label)
    tlog(f"using concurrency {concurrency}")
    folder = (
        config.tradeplan_folder
        if config.tradeplan_folder[-1] == "/"
        else f"{config.tradeplan_folder}/"
    )
    fname = f"{folder}{config.configuration_filename}"
    conf_dict = load_configuration(fname)

    if "optimizer" not in conf_dict:
        tlog(f"optimizer directive not found in {fname}. aborting.")
        sys.exit(0)

    tlog(f"optimizer directive found in {fname}")

    try:
        start_date = dateFromString(conf_dict["optimizer"]["start_date"])
        end_date = dateFromString(conf_dict["optimizer"]["end_date"])
    except Exception:
        tlog("failed to load start & end dates")
        sys.exit(0)

    optimize(
        start_date=start_date,
        end_date=end_date,
        hyperparameters=create_hyperparameters(
            conf_dict["optimizer"]["hyperparameters"]
        ),
        parameters=create_parameters(conf_dict["optimizer"]["parameters"]),
        conf_dict=conf_dict,
        concurrency=concurrency,
    )

    sys.exit(0)



import asyncio
import uuid

import fire
from tabulate import tabulate

from liualgotrader.analytics import analysis
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.types import AssetType
from liualgotrader.models.portfolio import Portfolio
from liualgotrader.reprocess.portfolio import account_transactions


def equity(portfolio_id: str):
    """Show current equity breakdown"""
    portfolio = asyncio.run(analysis.get_portfolio_equity(portfolio_id))
    if portfolio.empty:
        print("Empty portfolio")
        return

    print(
        tabulate(
            portfolio,
            headers=["symbol", "qty", "price ($)", "total ($)"],
            numalign="decimal",
            tablefmt="pretty",
            floatfmt=".2f",
            showindex=False,
        )
    )
    print(f"Number of stocks: {len(portfolio)}")


def account(portfolio_id: str):
    """Show account transactions for portfolio"""
    portfolio = asyncio.run(analysis.get_portfolio_cash(portfolio_id))
    if portfolio.empty:
        print("Empty portfolio transactions")
        return

    print(
        tabulate(
            portfolio,
            numalign="decimal",
            tablefmt="pretty",
            headers=["when", "amount ($)"],
            floatfmt=".2f",
        )
    )
    print(f"Number of transactions: {len(portfolio)}")


def recalc(portfolio_id):
    """Re-calculate portfolio's account transactions"""
    account_transactions(portfolio_id)


def create(account_size: float, credit_line: float, asset_type: str):
    """Create a new account"""
    portfolio_id = str(uuid.uuid4())
    try:
        asset = AssetType[asset_type]
    except KeyError:
        print(f"ASSET_TYPE {asset_type} not supported. aborting.")
        return

    asyncio.run(
        Portfolio.save(
            portfolio_id=portfolio_id,
            portfolio_size=account_size,
            credit=credit_line,
            parameters={},
            asset_type=asset,
        )
    )
    print(f"Portfolio ID {portfolio_id} created")


def trades(portfolio_id):
    """Display portfolio trades"""
    portfolio = analysis.load_trades_by_portfolio(portfolio_id)
    portfolio = (
        portfolio[["tstamp", "symbol", "operation", "qty", "price"]]
        .set_index("tstamp")
        .sort_index()
    )
    portfolio["total"] = portfolio.qty * portfolio.price
    portfolio = portfolio.round(2)
    print(
        tabulate(
            portfolio,
            numalign="decimal",
            tablefmt="pretty",
            headers=["when", "symbol", "op", "qty", "price ($)", "total ($)"],
            floatfmt=".2f",
        )
    )
    print(f"Number of trades: {len(portfolio)}")


def list():
    """List active portfolios"""

    data = []
    loop = asyncio.get_event_loop()
    loop.run_until_complete(create_db_connection())
    print("Portfolio(s):")
    data = loop.run_until_complete(Portfolio.list_portfolios())
    data = data[
        [
            "portfolio_id",
            "tstamp",
            "last_transaction",
            "size",
            "assets",
            "parameters",
        ]
    ]

    print(
        tabulate(
            data,
            headers=[
                "Id",
                "Created",
                "Last Used",
                "Size",
                "Asset Type",
                "Parameters",
            ],
            numalign="decimal",
            tablefmt="pretty",
            floatfmt=".2f",
            showindex=False,
        )
    )


def main_cli() -> None:
    fire.Fire(
        {
            "create": create,
            "equity": equity,
            "recalc": recalc,
            "account": account,
            "trades": trades,
            "list": list,
        },
        name="portfolio",
    )



import asyncio
import traceback
import uuid
from datetime import date
from typing import Dict

import liualgotrader
from liualgotrader import enhanced_backtest
from liualgotrader.common import config
from liualgotrader.common.hyperparameter import Hyperparameters
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import AssetType, TimeScale
from liualgotrader.models.optimizer import OptimizerRun


async def save_optimizer_session(
    optimizer_session_id: str, batch_id: str, hypers: Hyperparameters
):
    await OptimizerRun.save(optimizer_session_id, batch_id, str(hypers))


def backtest_iteration(
    optimizer_session_id: str,
    start_date: date,
    end_date: date,
    conf_dict: Dict,
    hypers: Hyperparameters,
):
    tlog(
        f"starting backtest with start_date={start_date}, end_date={end_date} w/ configuration={conf_dict}"
    )
    uid = str(uuid.uuid4())

    config.build_label = liualgotrader.__version__ if hasattr(liualgotrader, "__version__") else ""  # type: ignore
    try:
        if not asyncio.get_event_loop().is_closed():
            asyncio.get_event_loop().close()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(asyncio.new_event_loop())
        loop.run_until_complete(
            enhanced_backtest.backtest_main(
                uid,
                start_date,
                end_date,
                TimeScale.day,
                conf_dict,
                0.0,
                0.0,
                AssetType.US_EQUITIES,
            )
        )
    except KeyboardInterrupt:
        tlog("backtest() - Caught KeyboardInterrupt")
    except Exception as e:
        tlog(
            f"backtest() - exception of type {type(e).__name__} with args {e.args}"
        )
        traceback.print_exc()
    else:
        loop.run_until_complete(
            save_optimizer_session(optimizer_session_id, uid, hypers)
        )
    finally:
        print("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=")
        print(f"new batch-id: {uid}")



import asyncio
import getopt
import os
import pathlib
import sys
import time
import uuid

import pygit2
import requests

from liualgotrader.common import config
from liualgotrader.common.database import create_db_connection
from liualgotrader.models.accounts import Accounts
from liualgotrader.models.portfolio import Portfolio


def show_version():
    print(f"Liu Algo Trading Framework v{config.build_label}")


def show_usage():
    show_version()

    print()
    print("usage:")
    print("\t liu quickstart")
    print("\t OR")
    print("\t liu create portfolio <amount> [--credit=<credit>]")
    print("\n")
    print("options:")
    print("\t--credit\tAccount credit line")
    print()


def setup_db(
    db_location: str,
    db_name: str,
    user_name: str,
    password: str,
    restore_sample_db: bool,
):
    try:
        print()
        print("+---------------------+")
        print("| setting up database |")
        print("+---------------------+")
        print()

        p = pathlib.Path(db_location)
        if not p.exists():
            p.mkdir()
        if db_location[-1] != "/":
            db_location += "/"

        print("downloading files..")
        if os.name == "nt":
            r = requests.get(
                "https://raw.github.com/amor71/LiuAlgoTrader/master/database/docker-compose-windows.yml"
            )
        else:
            r = requests.get(
                "https://raw.github.com/amor71/LiuAlgoTrader/master/database/docker-compose.yml"
            )
        resolved = (
            r.text.replace("{db_name}", db_name)
            .replace("{user_name}", user_name)
            .replace("{password}", password)
        )

        with open(f"{db_location}docker-compose.yml", "w") as f:
            f.write(resolved)
            f.write("\n")

        base_url = (
            "https://raw.github.com/amor71/LiuAlgoTrader/master/database/"
        )
        files = ["schema.sql", "postgres.conf"]
        if restore_sample_db:
            files.append("liu_dump.sql")

        for file in files:
            r = requests.get(f"{base_url}{file}")
            with open(f"{db_location}{file}", "w", encoding="utf-8") as f:
                f.write(r.text)
                f.write("")
    except Exception as e:
        print(f"Something went wrong:{e}")
        print(
            "Pls open a new issue w/ details 'https://github.com/amor71/LiuAlgoTrader/issues/new'"
        )
        exit(0)

    print("files loaded successfully.")
    print(f"running docker-compose from within {db_location}:")
    to_run = f"cd {db_location} && docker-compose up -d"
    print("> ", to_run)
    os.system(to_run)

    if restore_sample_db:
        print("waiting for database to complete setup. 2 minutes")
        time.sleep(120)
        print("restoring sample database")
        to_run = f"cd {db_location} && docker exec -i pg-docker psql -q -U liu -W liu liu < liu_dump.sql"
        print("> ", to_run)
        os.system(to_run)
    print()
    print("check deployment using `\psql -h localhost -p 5400 -U liu`")


def setup_samples(
    samples_location: str, user: str, passwd: str, db: str
) -> None:
    try:
        print()
        print("+--------------------+")
        print("| setting up samples |")
        print("+--------------------+")

        p = pathlib.Path(samples_location)
        if not p.exists():
            p.mkdir()
        if samples_location[-1] != "/":
            samples_location += "/"

        base_url = "https://raw.github.com/amor71/LiuAlgoTrader/master/examples/quickstart/"
        files = [
            "tradeplan.toml",
            "vwap_short.py",
            "momentum_long_simplified.py",
        ]

        for file in files:
            print(f"Downloading {base_url}{file} to {samples_location}...")
            r = requests.get(f"{base_url}{file}")
            with open(f"{samples_location}{file}", "w") as f:
                f.write(r.text)
                f.write("")

        print()
        print("creating environment variables script")

        if os.name == "nt":
            with open(f"{samples_location}env_vars.bat", "w") as f:
                f.write(
                    f"set DSN=postgresql://{user}:{passwd}@localhost:5400/{db}\n"
                )
        else:
            with open(f"{samples_location}env_vars.sh", "w") as f:
                f.write(
                    f"export DSN=postgresql://{user}:{passwd}@localhost:5400/{db}\n"
                )

    except Exception as e:
        print(f"Something went wrong:{e}")
        print(
            "Pls open a new issue w/ details 'https://github.com/amor71/LiuAlgoTrader/issues/new'"
        )
        exit(0)


def quickstart():
    print(f"Welcome to Lig Algo Trading Framework v{config.build_label}!")
    print()
    print("This wizard will guide you through the setup process.")
    print()
    print("+----------------------------------------+")
    print("| Step #1 - Alpaca & Polygon credentials |")
    print("+----------------------------------------+")
    print()
    print(
        "To use Liu Algo Trading Framework you need an account with Alpaca Markets,"
    )
    print("do you already have an account [Y]/n:")
    i = input()
    have_funded = len(i) == 0 or (i == "y" or i == "Y" or i.lower() == "yes")

    if not have_funded:
        print("For additional details `https://alpaca.markets/docs/about-us/`")
        return

    to_exit = False
    if not config.alpaca_api_key or not config.alpaca_api_secret:
        print(
            "Liu Algo Trading Framework uses Alpaca for both LIVE and PAPER trading."
        )
        print()
        print("The Framework expects three environment variables to be set:")
        print(
            "`APCA_API_KEY_ID` and `APCA_API_SECRET_KEY`: reflecting the funded"
        )
        print("account's API key and secret respectively.")
        print(
            "And `APCA_API_BASE_URL` reflecting the base URL for your account."
        )
        print(
            "Please set the three environment variables and re-run the wizard."
        )
        to_exit = True

    if to_exit:
        sys.exit(0)

    print()
    print("+----------------------------------+")
    print("| Step #2 - Database configuration |")
    print("+----------------------------------+")
    print()
    print("Do you already have a PostgreSQL instance configured [N]/y:")
    i = input()
    already_have_db = len(i) > 0 and (
        i == "y" or i == "Y" or i.lower() == "yes"
    )

    if already_have_db:
        print(
            "Follow the instructions at 'https://liualgotrader.readthedocs.io/en/latest/(Advanced)%20Setup.html#database-setup' to complete your database setup."
        )
        restore_sample_db = False
    else:
        pwd = pathlib.Path().absolute()
        print(
            "Liu Algo Trading Framework uses `docker-compose` to run a local database."
        )
        print(
            "The installation will download the docker-compose.yml, database schema"
        )
        print(
            "and prepare the database for first time usage. You can stop and re-run"
        )
        print(
            "the database using `docker-compose up -d ` and `docker-compose down`"
        )
        print(
            "respectively. Your data will not be deleted. For more details RTFM."
        )
        print()
        print(f"Select location for database files [{pwd}/liu_data/]:")
        db_location = input()
        db_location = (
            f"{pwd}/liu_data/" if not len(db_location) else db_location
        )

        print()
        print("** IMPORTANT NOTE**")
        print(
            "The installation wizard, aside from installing PostgreSQL in a docker"
        )
        print(
            "may also download a sample database, with existing data to help you"
        )
        print("take your first steps with Liu Algo Trading Framework.")
        print(
            "However, in order to restore the sample data, you need to select"
        )
        print(
            "the default username ('liu') password ('liu)', and database name ('liu')."
        )
        print(
            "after you try out the sample, you can delete data and re-run this"
        )
        print("wizard to install a fresh copy of the database.")
        print()
        print("Would you like to download the sample database? [Y]/n")
        i = input()
        restore_sample_db = len(i) == 0 or (
            i == "y" or i == "Y" or i.lower() == "yes"
        )
        print(
            "Select the database name for keeping track of your trading [liu]:"
        )
        db_name = input()
        db_name = "liu" if not len(db_name) else db_name
        print("Select the database user-name [liu]:")
        user_name = input()
        user_name = "liu" if not len(user_name) else user_name
        print("Select the database password [liu]:")
        password = input()
        password = "liu" if not len(password) else password

    print()
    print("+--------------------+")
    print("| Step #3 - Examples |")
    print("+--------------------+")
    print()
    print("Would you like to download & view samples? [Y]/n:")
    i = input()
    samples = not len(i) or i.lower() in ("y", "yes")
    sample_location = None
    if samples:
        pwd = pathlib.Path().absolute()
        print(f"Select location for sample files [{pwd}/liu_samples]:")
        sample_location = input()
        sample_location = (
            f"{pwd}/liu_samples"
            if not len(sample_location)
            else sample_location
        )

    print()
    print("Ready to go?? Press [ENTER] to start the installation..")
    input()

    if not already_have_db:
        setup_db(db_location, db_name, user_name, password, restore_sample_db)
    if samples:
        setup_samples(sample_location, user_name, password, db_name)
    print("setup completed successfully.")
    print()
    print()
    print("+---------------+")
    print("| what's next?  |")
    print("+---------------+")
    print()
    print("Congratulations, Liu Algo Trader is ready to go!")
    print(
        "From here, you can either run the back-testing UI, or read the docs."
    )
    print(
        "The full documentation is available here: `https://liualgotrader.rtfd.io`."
    )
    if restore_sample_db:
        print()
        print(
            "The restored database includes data for the first week and half of Oct 2020."
        )
        print(
            "You may now run a back-testing sessions for these days, as well as "
        )
        print("analyze an existing trading session.")
    print()

    if sample_location:
        print("To experience Liu Algo Trading for the first time:")
        print(
            f"1. change your working directory to the sample directory {sample_location} ('cd {sample_location}'),"
        )
        if os.name == "nt":
            print("2. run the env-vars script ('env_vars.bat'),")
        else:
            print("2. run the env-vars script ('source env_vars.sh'),")
        print(
            "3.`streamlit run https://raw.github.com/amor71/LiuAlgoTrader/master/analysis/backtester_ui.py`"
        )
        if restore_sample_db:
            print("4. On the app selection select the 'analyzer` app")
            print(
                "5. copy and paste the batch-id '2398380c-5146-4b58-843a-a50c458c8071' and press <enter>"
            )

    print()
    print("Thank you for choosing Liu Algo Trading Framework. ")
    print()


def create_account(amount: float, credit: float):
    loop = asyncio.get_event_loop()
    loop.run_until_complete(create_db_connection())
    account_id = loop.run_until_complete(
        Accounts.create(amount, allow_negative=credit > 0, credit_line=credit)
    )
    print(f"Account ID {account_id} created")


def create_portfolio(amount: float, credit: float):
    portfolio_id = str(uuid.uuid4())
    loop = asyncio.get_event_loop()
    loop.run_until_complete(create_db_connection())
    loop.run_until_complete(
        Portfolio.save(
            portfolio_id=portfolio_id,
            portfolio_size=amount,
            credit=credit,
            parameters={},
        )
    )
    print(f"Portfolio ID {portfolio_id} created")


def main_cli() -> None:
    config.filename = os.path.basename(__file__)

    try:
        config.build_label = pygit2.Repository("../").describe(
            describe_strategy=pygit2.GIT_DESCRIBE_TAGS
        )
    except pygit2.GitError:
        import liualgotrader

        config.build_label = liualgotrader.__version__ if hasattr(liualgotrader, "__version__") else ""  # type: ignore

    if len(sys.argv) < 2:
        show_usage()
        exit(0)

    if sys.argv[1] == "quickstart":
        try:
            quickstart()
        except KeyboardInterrupt:
            print("Oops... exiting gracefully")
    elif sys.argv[1] == "create":
        try:
            if sys.argv[2] == "portfolio":
                if float(sys.argv[3]) <= 0.0:
                    print(
                        "Oops.. account size must be positive. Always think positive w/ Liu"
                    )
                    exit(0)
                amount = float(sys.argv[3])

                credit = 0.0
                opts, args = getopt.getopt(
                    sys.argv[4:],
                    shortopts="",
                    longopts=[
                        "credit=",
                    ],
                )
                for opt, arg in opts:
                    if opt in ("--credit"):
                        credit = float(arg)
                        if credit <= 0.0:
                            print(
                                "Oops.. creditamount must be positive. Always think positive w/ Liu"
                            )
                            exit(0)

                create_portfolio(amount, credit)
        except Exception as e:
            print(f"[ERROR] {e}")
            show_usage()
    else:
        show_usage()

    exit(0)



import copy
import multiprocessing as mp
import os
import sys
import uuid
from multiprocessing import Queue
from typing import List

import pygit2
import toml
from pytz import timezone

from liualgotrader.common import config
from liualgotrader.common.concurrency import calc_num_consumer_processes
from liualgotrader.common.tlog import tlog
from liualgotrader.consumer import consumer_main
from liualgotrader.producer import producer_main
from liualgotrader.scanners_runner import main


def motd(filename: str, version: str, unique_id: str) -> None:
    """Display welcome message"""

    tlog("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
    tlog("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
    tlog(f"{filename} {version} starting")
    tlog(f"unique id: {unique_id}")
    tlog("----------------------------------------------------------")
    tlog("----------------------------------------------------------")


def main_cli():
    config.filename = os.path.basename(__file__)
    mp.set_start_method("spawn")

    try:
        config.build_label = pygit2.Repository("../").describe(
            describe_strategy=pygit2.GIT_DESCRIBE_TAGS
        )
    except pygit2.GitError:
        import liualgotrader

        config.build_label = liualgotrader.__version__ if hasattr(liualgotrader, "__version__") else ""  # type: ignore

    uid = str(uuid.uuid4())
    motd(filename=config.filename, version=config.build_label, unique_id=uid)

    # load configuration
    folder = (
        config.tradeplan_folder
        if config.tradeplan_folder[-1] == "/"
        else f"{config.tradeplan_folder}/"
    )
    fname = f"{folder}{config.configuration_filename}"
    try:
        conf_dict = copy.deepcopy(toml.load(fname))
        print(conf_dict)
        tlog(f"loaded configuration file from {fname}")
    except FileNotFoundError:
        tlog(f"[ERROR] could not locate tradeplan file {fname}")
        sys.exit(0)

    # Consumers first
    num_consumer_processes = calc_num_consumer_processes()
    tlog(f"Starting {num_consumer_processes} consumer processes")
    queues: List[Queue] = [Queue() for _ in range(num_consumer_processes)]
    consumers = [
        mp.Process(
            target=consumer_main,
            args=(queues[i], uid, conf_dict),
        )
        for i in range(num_consumer_processes)
    ]
    for p in consumers:
        # p.daemon = True
        p.start()

    scanner_queue: mp.Queue = mp.Queue()
    producer = mp.Process(
        target=producer_main,
        args=(
            uid,
            queues,
            conf_dict,
            scanner_queue,
            num_consumer_processes,
        ),
    )
    producer.start()

    scanner = None
    if "scanners" in conf_dict:
        tlog("Starting scanners process")
        scanner = mp.Process(
            target=main,
            args=(
                conf_dict,
                scanner_queue,
            ),
        )
        scanner.start()

    # wait for completion and hope everyone plays nicely
    try:
        producer.join()
        if scanner:
            scanner.join()

        for p in consumers:
            p.join()

    except KeyboardInterrupt:
        producer.kill()

        if scanner:
            scanner.terminate()

        for p in consumers:
            p.terminate()

    sys.exit(0)



"""off-hours calculations, and data collections"""
import asyncio
import importlib.util
import os
import sys
import traceback
from typing import Dict, List, Optional

import pygit2
import toml

from liualgotrader.common import config
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.tlog import tlog
from liualgotrader.miners.base import Miner

# rom liualgotrader.miners.stock_cluster import StockCluster


def motd(filename: str, version: str) -> None:
    """Display welcome message"""

    print("+=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=+")
    tlog(f"{filename} {version} starting")
    print("+=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=+")
    tlog(f"DSN: {config.dsn}")
    print("+=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=+")


async def main(conf_dict: Dict):
    task_list: List[Optional[asyncio.Task]] = []

    await create_db_connection()
    for miner in conf_dict["miners"]:
        try:
            if "filename" in conf_dict["miners"][miner]:
                spec = importlib.util.spec_from_file_location(
                    "module.name", conf_dict["miners"][miner]["filename"]
                )
                if not spec:
                    raise AssertionError(
                        f"could not load module {conf_dict['miners'][miner]['filename']}"
                    )
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)  # type: ignore
            else:
                module = importlib.import_module(
                    f"liualgotrader.miners.{miner}"
                )
            class_name = f"{miner[0].upper()}{miner[1:]}"
            miner_class = getattr(module, class_name)

            if not issubclass(miner_class, Miner):
                tlog(f"Miner must inherit from class {Miner.__name__}")
                exit(0)
        except Exception as e:
            tlog(f"[ERROR] miner {miner} resulted in exception:`{e}`")


def main_cli() -> None:
    """
    starting
    """
    try:
        build_label = pygit2.Repository("../").describe(
            describe_strategy=pygit2.GIT_DESCRIBE_TAGS
        )
    except pygit2.GitError:
        import liualgotrader

        build_label = liualgotrader.__version__ if hasattr(liualgotrader, "__version__") else ""  # type: ignore
    config.build_label = build_label
    filename = os.path.basename(__file__)
    motd(filename=filename, version=build_label)

    # load configuration
    tlog(
        f"loading configuration file from {os.getcwd()}/{config.miner_configuration_filename}"
    )
    try:
        conf_dict = toml.load(config.miner_configuration_filename)
    except FileNotFoundError:
        tlog(
            f"[ERROR] could not locate market_miner configuration file {config.miner_configuration_filename}"
        )
        sys.exit(0)

    try:
        if not asyncio.get_event_loop().is_closed():
            asyncio.get_event_loop().close()
        asyncio.run(main(conf_dict))
    except KeyboardInterrupt:
        tlog("market_miner.main() - Caught KeyboardInterrupt")
    except Exception as e:
        tlog(
            f"market_miner.main() - exception of type {type(e).__name__} with args {e.args}"
        )
        exc_info = sys.exc_info()
        traceback.print_exception(*exc_info)
        del exc_info

    tlog("*** market_miner completed ***")



import getopt
import os
import sys
from datetime import date, datetime
from typing import List, Optional

import pandas as pd
import parsedatetime.parsedatetime as pdt
import pygit2
import pytz
import toml
from requests.exceptions import HTTPError

from liualgotrader import enhanced_backtest
from liualgotrader.common import config, market_data, trading_data
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.decorators import timeit
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import AssetType, TimeScale
from liualgotrader.fincalcs.vwap import add_daily_vwap
from liualgotrader.models.algo_run import AlgoRun
from liualgotrader.models.new_trades import NewTrade
from liualgotrader.models.trending_tickers import TrendingTickers
from liualgotrader.scanners.base import Scanner
from liualgotrader.scanners.momentum import Momentum
from liualgotrader.strategies.base import Strategy, StrategyType


def show_usage():
    print(
        f"\n{sys.argv[0]} from <start_date> [--asset=equity(DEFAULT)|crypto][--scanners=<scanner-name,>] [--strats=<strategy-name,>] [--to=<end_date> DEFAULT is today] [--scale=day(DEFAULT)|minute [--buy-fee-percentage=0.(DEFAULT)] [--sell-fee-percentage=0.(DEFAULT)]",
    )

    print("\n\noptions:")
    print(
        "asset\t\t\tAsset type being traded. equity = US Equities, crypt = Crypto-currency. Asset type affects the market schedule for backtesting."
    )
    print(
        "to\t\t\tdate string in the format YYYY-MM-DD, if not provided current day is selected"
    )
    print("scale\t\t\ttime-scale for loading past data for back-test-ing")
    print(
        "buy-fee-percentage\tBroker fees as percentage from transaction. Represented as 0-1."
    )
    print(
        "sell-fee-percentage\tBroker fees as percentage from transaction. Represented as 0-1."
    )


def show_version(filename: str, version: str) -> None:
    """Display welcome message"""
    print(f"filename:{filename}\ngit version:{version}\n")


def dateFromString(s: str) -> date:
    c = pdt.Calendar()
    result, what = c.parse(s)
    dt = None

    # what was returned (see http://code-bear.com/code/parsedatetime/docs/)
    # 0 = failed to parse
    # 1 = date (with current time, as a struct_time)
    # 2 = time (with current date, as a struct_time)
    # 3 = datetime
    if what in (1, 2):
        # result is struct_time
        dt = datetime(*result[:6]).date()
    elif what == 3:
        # result is a datetime
        dt = result.date()

    if dt is None:
        raise ValueError(f"Don't understand date '{s}'")

    return dt


def main_cli() -> None:
    try:
        config.build_label = pygit2.Repository("../").describe(
            describe_strategy=pygit2.GIT_DESCRIBE_TAGS
        )
    except pygit2.GitError:
        import liualgotrader

        config.build_label = liualgotrader.__version__ if hasattr(liualgotrader, "__version__") else ""  # type: ignore

    if len(sys.argv) == 1:
        show_usage()
        sys.exit(0)

    config.filename = os.path.basename(__file__)

    folder = (
        config.tradeplan_folder
        if config.tradeplan_folder[-1] == "/"
        else f"{config.tradeplan_folder}/"
    )
    fname = f"{folder}{config.configuration_filename}"
    try:
        conf_dict = toml.load(fname)
        tlog(f"loaded configuration file from {fname}")
    except FileNotFoundError:
        tlog(f"[ERROR] could not locate tradeplan file {fname}")
        sys.exit(0)
    conf_dict = toml.load(config.configuration_filename)
    config.portfolio_value = conf_dict.get("portfolio_value", None)
    if "risk" in conf_dict:
        config.risk = conf_dict["risk"]

    if sys.argv[1] == "from":
        from_date = dateFromString(sys.argv[2])
        tlog(f"selected {sys.argv[2]} applied {from_date}")
        try:
            scanners: Optional[List] = None
            strategies: Optional[List] = None
            to_date = datetime.now(tz=pytz.timezone("US/Eastern")).date()
            scale = TimeScale["day"]
            buy_fee_percentage = 0.0
            sell_fee_percentage = 0.0
            asset_type = AssetType.US_EQUITIES
            opts, args = getopt.getopt(
                sys.argv[3:],
                shortopts="",
                longopts=[
                    "to=",
                    "scale=",
                    "scanners=",
                    "strats=",
                    "buy-fee-percentage=",
                    "sell-fee-percentage=",
                    "asset=",
                ],
            )
            for opt, arg in opts:
                if opt in ("--scanners"):
                    scanners = arg.split(",")
                elif opt in ("--strats"):
                    strategies = arg.split(",")
                elif opt in ("--to"):
                    to_date = dateFromString(arg)
                elif opt in ("--buy-fee-percentage"):
                    buy_fee_percentage = float(arg)
                elif opt in ("--sell-fee-percentage"):
                    sell_fee_percentage = float(arg)
                elif opt in ("--asset"):
                    if arg.lower() == "crypto":
                        asset_type = AssetType.CRYPTO
                    elif arg.lower() == "equity":
                        asset_type = AssetType.US_EQUITIES
                    else:
                        print(
                            f"ERROR: value {arg} not supported for parameter 'asset'"
                        )
                        sys.exit(0)
                elif opt in ("--scale"):
                    if arg in ("day", "minute"):
                        scale = TimeScale[arg]
                    else:
                        print(
                            f"ERROR: wrong `scale` parameter {arg} passed: not implemented yet"
                        )
                        sys.exit(0)
        except getopt.GetoptError as e:
            print(f"Error parsing options:{e}\n")
            show_usage()
            sys.exit(0)

        enhanced_backtest.backtest(
            from_date=from_date,
            to_date=to_date,
            scale=scale,
            config=conf_dict,
            scanners=scanners,
            strategies=strategies,
            asset_type=asset_type,
            buy_fee_percentage=buy_fee_percentage,
            sell_fee_percentage=sell_fee_percentage,
        )
    else:
        print(f"Error parsing {sys.argv[1:]}\n")
        show_usage()
        sys.exit(0)

    sys.exit(0)
